#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 13.02.2023
# Version: 2.0.0
#
# Description:
# This module contains the functions for performing file operations for the moosez.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez to perform file operations.
#
# ----------------------------------------------------------------------------------------------------------------------

import os
from datetime import datetime
from typing import Union, Tuple, List
from moosez import constants


def create_directory(directory_path: str) -> None:
    """
    Creates a directory at the specified path.
    
    :param directory_path: The path to the directory.
    :type directory_path: str
    """
    if not os.path.isdir(directory_path):
        os.makedirs(directory_path)


def get_files(directory: str, prefix: Union[str, Tuple[str, ...]], suffix: Union[str, Tuple[str, ...]]) -> List[str]:
    """
    Returns the list of files in the directory with the specified wildcard.
    :param directory: The directory path.
    :type directory: str
    :param suffix: valid suffixes.
    :type suffix: str
    :param prefix: valid prefixes.
    :type prefix: str
    :return: The list of files.
    :rtype: list
    """

    if isinstance(prefix, str):
        prefix = (prefix,)

    if isinstance(suffix, str):
        suffix = (suffix,)

    files = [file for file in os.listdir(directory) if os.path.isfile(os.path.join(directory, file))]

    modality_files = []
    for file in files:
        if file.startswith(prefix) and file.endswith(suffix):
            modality_files.append(os.path.join(directory, file))
    return modality_files


def get_modality_file(directory: str, modality_prefix: str) -> Union[str, None]:
    """
    Finds the modality file in the specified folder.
    :param directory: The path to the folder.
    :type directory: str
    :param modality_prefix: The modality prefix.
    :type modality_prefix: str
    :return: The path to the modality file.
    :rtype: str
    """

    modality_files = get_files(directory, modality_prefix, ('.nii', '.nii.gz'))

    if len(modality_files) == 1:
        return modality_files[0]
    elif len(modality_files) > 1:
        raise ValueError("More than one modality file found in the directory.")
    else:
        return None


def moose_folder_structure(parent_directory: str) -> Tuple[str, str, str]:
    """
    Creates the moose folder structure.
    
    :param parent_directory: The path to the parent directory.
    :type parent_directory: str
    
    :return: A tuple containing the paths to the moose directory, output directory, and stats directory.
    :rtype: tuple
    """
    moose_dir = os.path.join(parent_directory, 'moosez-' + datetime.now().strftime('%Y-%m-%d-%H-%M-%S'))
    create_directory(moose_dir)

    segmentation_dir = os.path.join(moose_dir, constants.SEGMENTATIONS_FOLDER)
    stats_dir = os.path.join(moose_dir, constants.STATS_FOLDER)
    create_directory(segmentation_dir)
    create_directory(stats_dir)
    return moose_dir, segmentation_dir, stats_dir


def get_nifti_file_stem(file_path: str) -> str:
    file_stem = os.path.basename(file_path).split('.gz')[0].split('.nii')[0]
    return file_stem



"""Testing End 2 End."""
from pathlib import Path
from moosez import moose

data = Path(__file__).parent / "data"

def test_basic_e2e(tmp_path) -> None:
    min_ct_throat = data / "CT_enhance_pet_1446_neck_slice.nii.gz"
    moose(input_data=str(min_ct_throat),
          model_names="clin_ct_organs",
          output_dir=tmp_path,
          accelerator="cpu")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import torch
import os
import emoji
import pyfiglet
import platform
import json
import importlib.metadata
from halo import Halo
from datetime import datetime
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from rich.console import Console, RenderableType
from rich.table import Table
from rich.panel import Panel
from rich.text import Text
from rich.progress import Progress, TextColumn, BarColumn, FileSizeColumn, TransferSpeedColumn, TimeRemainingColumn
from typing import Union, Tuple, List, Dict, Optional
from moosez.constants import ANSI_VIOLET, ANSI_RESET


# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 13.02.2023
# Version: 2.0.0
#
# Description:
# This module contains the urls and filenames of the models and binaries that are required for the moosez.
#
# Usage:
# The variables in this module can be imported and used in other modules within the moosez to download the necessary
# binaries and models for the moosez.
#
# ----------------------------------------------------------------------------------------------------------------------


class OutputManager:
    def __init__(self, verbose_console: bool, verbose_log: bool):
        self.verbose_console = verbose_console
        self.verbose_log = verbose_log

        self.console = Console(quiet=not self.verbose_console)
        self.spinner = Halo(spinner='dots', enabled=self.verbose_console)
        self.spinner_running = False

        self.logger = None
        self.nnunet_log_filename = os.devnull

        # On Windows, the console encoding may not be UTF-8, which can cause issues with certain characters
        self.sanitize_console_encoding = self.console.encoding.lower() not in ['utf-8', 'utf8']

    def create_file_progress_bar(self):
        progress_bar = Progress(TextColumn("[bold blue]{task.description}"), BarColumn(bar_width=None),
                                "[progress.percentage]{task.percentage:>3.0f}%", "•", FileSizeColumn(),
                                TransferSpeedColumn(), TimeRemainingColumn(), console=self.console, expand=True)
        return progress_bar

    def create_progress_bar(self):
        progress_bar = Progress(console=self.console)
        return progress_bar

    def create_table(self, header: List[str], styles: Union[List[str], None] = None) -> Table:
        table = Table()
        if styles is None:
            styles = [None] * len(header)
        for header, style in zip(header, styles):
            table.add_column(header, style = style)
        return table

    def configure_logging(self, log_file_directory: Union[str, None]):
        if not self.verbose_log or self.logger:
            return

        if log_file_directory is None:
            log_file_directory = os.getcwd()

        timestamp = datetime.now().strftime('%H-%M-%d-%m-%Y')

        self.nnunet_log_filename = os.path.join(log_file_directory, f'moosez-v{MOOSE_VERSION}_nnUNet_{timestamp}.log')

        self.logger = logging.getLogger(f'moosez-v{MOOSE_VERSION}')
        self.logger.setLevel(logging.INFO)
        self.logger.propagate = False

        if not any(isinstance(handler, logging.FileHandler) for handler in self.logger.handlers):
            log_format = '%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s'
            formatter = logging.Formatter(log_format)

            log_filename = os.path.join(log_file_directory, f'moosez-v{MOOSE_VERSION}_{timestamp}.log')
            file_handler = logging.FileHandler(log_filename, mode='w')
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)

            self.logger.addHandler(file_handler)

    def log_update(self, text: str):
        if self.verbose_log and self.logger:
            self.logger.info(text)

    def console_update(self, text: Union[str, RenderableType]):
        if isinstance(text, str):
            text = Text.from_ansi(text)

        if self.sanitize_console_encoding and isinstance(text, Text):
            # Ensure the text can be represented in the console by robust encoding and decoding (ignoring invalid characters)
            text = Text(text.plain.encode(self.console.encoding, "ignore").decode(self.console.encoding))

        self.console.print(text)

    def spinner_update(self, text: str):
        if self.spinner.enabled:
            if not self.spinner_running:
                self.spinner.start()
                self.spinner_running = True
            self.spinner.text = text

    def spinner_stop(self):
        if self.spinner.enabled and self.spinner_running:
            self.spinner.stop()
            self.spinner_running = False

    def spinner_start(self, text: str):
        if self.spinner.enabled and not self.spinner_running:
            self.spinner.start(text)
            self.spinner_running = True

    def spinner_succeed(self, text: str):
        if self.spinner.enabled:
            self.spinner.succeed(text)
            self.spinner_running = False

    def spinner_warn(self, text: str):
        if self.spinner.enabled:
            self.spinner.warn(text)
            self.spinner_running = False

    @contextmanager
    def manage_nnUNet_output(self):
        target_path = self.nnunet_log_filename if self.verbose_log else os.devnull
        mode = "a" if self.verbose_log else 'w'
        with open(target_path, mode) as target, redirect_stdout(target), redirect_stderr(target):
            yield

    def display_logo(self):
        """
        Display MOOSE logo

        This function displays the MOOSE logo using the pyfiglet library and ANSI color codes.

        :return: None
        """
        self.console_update(' ')
        result = ANSI_VIOLET + pyfiglet.figlet_format(f" MOOSE {MOOSE_VERSION}", font="smslant").rstrip() + ANSI_RESET
        text = ANSI_VIOLET + " A part of the ENHANCE community. Join us at www.enhance.pet to build the future of" \
                             " PET imaging together." + ANSI_RESET
        self.console_update(result)
        self.console_update(text)
        self.console_update(' ')

    def display_authors(self):
        """
        Display manuscript citation

        This function displays authors for the MOOSE project.

        :return: None
        """
        self.console_update(f'{ANSI_VIOLET} {emoji.emojize(":desktop_computer:")}  AUTHORS:{ANSI_RESET}')
        self.console_update(" ")
        self.console_update(" The Three Moose-keteers 🤺: Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer | Manuel Pires")
        self.console_update(" ")

    def display_doi(self):
        """
        Display manuscript citation

        This function displays the manuscript citation for the MOOSE project.

        :return: None
        """
        self.console_update(f'{ANSI_VIOLET} {emoji.emojize(":scroll:")} CITATION:{ANSI_RESET}')
        self.console_update(" ")
        self.console_update(" Fully Automated, Semantic Segmentation of Whole-Body [18F]-FDG PET/CT Images Based on Data-Centric Artificial Intelligence")
        self.console_update(" 10.2967/jnumed.122.264063")
        self.console_update(" ")
        self.console_update(" Copyright 2022, Quantitative Imaging and Medical Physics Team, Medical University of Vienna")

    def display_docker_usage(self, docker_bind: str, example_model: str):
        docker_example = f"""\
        docker run --gpus all --ipc=host \\
        -v /path/to/data:/app/data \\
        -v {docker_bind}:/usr/local/models \\
        lalithshiyam/moosez \\
        -d /app/data -m {example_model}
        """
        docker_msg = Text()
        docker_msg.append("🐳 Some useful information...\n", style="bold violet")
        docker_msg.append("\n📂 Downloaded models are stored here:\n", style="bold")
        docker_msg.append(f"  {docker_bind}/nnunet_trained_models\n\n", style="orange3")
        docker_msg.append("📦 Bind this folder into the container at:\n", style="bold")
        docker_msg.append("  /usr/local/models\n\n", style="orange3")
        docker_msg.append("🚀 Example Docker Run:\n", style="bold")
        docker_msg.append(docker_example, style="green")

        self.console.print(Panel(docker_msg, title="🐳 INFO FOR DOCKER USERS", border_style="violet", padding=(1, 2)))


def check_device(output_manager: Optional[OutputManager] = None) -> Tuple[str, Union[int, None]]:
    """
    This function checks the available device for running predictions, considering CUDA, MPS (for Apple Silicon) or CPU as fallback.
    """
    if output_manager is None:
        output_manager = OutputManager(False, False)

    accelerator_information = get_accelerator_information()
    accelerator = accelerator_information["accelerator"]

    if accelerator == "cuda":
        output_manager.console_update(f" CUDA is available with {accelerator_information['device_count']} GPU(s). Predictions will be run on GPU.")
        return "cuda", accelerator_information["device_count"]

    if accelerator == "mps":
        output_manager.console_update(" Apple MPS backend is available. Predictions will be run on Apple Silicon GPU.")
        return "mps", None

    output_manager.console_update(" CUDA/MPS not available. Predictions will be run on CPU.")
    return "cpu", None


def get_accelerator_information() -> Dict:
    if torch.cuda.is_available():
        return {"accelerator": "cuda",
                "cuda_version": torch.version.cuda,
                "device_count": torch.cuda.device_count(),
                "device_name": ", ".join([torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())])}
    elif torch.backends.mps.is_available():
        return {"accelerator": "mps"}
    else:
        return {"accelerator": "cpu"}


def get_system_information() -> Dict:
    metadata = {"tool_name": "moosez",
                "tool_version": MOOSE_VERSION,
                "timestamp_start": datetime.now().strftime('%d.%m.%Y %H:%M'),
                "python_version": platform.python_version(),
                "os": platform.system(),"platform": platform.platform(),
                "architecture": platform.machine()}
    metadata.update(get_accelerator_information())

    return metadata


def write_information_json(metadata: Dict, metadata_file_directory: str) -> None:
    metadata["timestamp_stop"] = datetime.now().strftime('%d.%m.%Y %H:%M')

    os.makedirs(metadata_file_directory, exist_ok=True)
    metadata_file_path = os.path.join(metadata_file_directory, "MOOSE_information.json")
    with open(metadata_file_path, "w") as f:
        json.dump(metadata, f, indent=4)


os.environ["nnUNet_raw"] = ""
os.environ["nnUNet_preprocessed"] = ""
os.environ["nnUNet_results"] = ""
try:
    MOOSE_VERSION = importlib.metadata.version("moosez")
except importlib.metadata.PackageNotFoundError:
    MOOSE_VERSION = "0.0.0"
MOOSE_ROOT_PATH: str = os.path.dirname(os.path.abspath(__file__))
MODELS_DIRECTORY_PATH: str = os.path.join(MOOSE_ROOT_PATH, 'models', 'nnunet_trained_models')



import os
from pathlib import Path
import requests
from typing import Union
from moosez import constants
from moosez import system
from moosez import file_utilities


def download_enhance_data(download_directory: Union[str, None], output_manager: system.OutputManager):
    output_manager.log_update(f"    - Downloading ENHANCE 1.6k data")
    if not download_directory:
        download_directory = get_default_download_folder()

    file_utilities.create_directory(download_directory)

    download_file_name = os.path.basename(constants.ENHANCE_URL)
    download_file_path = os.path.join(download_directory, download_file_name)

    response = requests.get(constants.ENHANCE_URL, stream=True)
    if response.status_code != 200:
        output_manager.console_update(f"    X Failed to download model from {constants.ENHANCE_URL}")
        raise Exception(f"Failed to download model from {constants.ENHANCE_URL}")
    total_size = int(response.headers.get("Content-Length", 0))
    chunk_size = 1024 * 10

    progress = output_manager.create_file_progress_bar()
    with progress:
        task = progress.add_task(f"[white] Downloading ENHANCE 1.6k data (Powered by AWS) ", total=total_size)
        with open(download_file_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)
                    progress.update(task, advance=chunk_size)
    output_manager.console_update(f"{constants.ANSI_GREEN} ENHANCE 1.6k data successfuly downloaded. {constants.ANSI_RESET}")


def get_default_download_folder():
    if os.name == 'nt':  # For Windows
        download_folder = Path(os.getenv('USERPROFILE')) / 'Downloads'
    else:  # For macOS and Linux
        download_folder = Path.home() / 'Downloads'

    return download_folder



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
#         Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 05.06.2023
# Version: 2.0.0
#
# Description:
# This module handles image processing for the moosez.
#
# Usage:
# The functions in this module can be imported and used in other modules within the moosez for image processing.
#
# ----------------------------------------------------------------------------------------------------------------------

import SimpleITK
import itertools
import dask.array as da
import numpy as np
import pandas as pd
import scipy.ndimage as ndimage
import nibabel
import os
import math
import tempfile
from typing import Union, Tuple, List, Dict
from moosez.constants import CHUNK_THRESHOLD_RESAMPLING, CHUNK_THRESHOLD_INFERRING
from moosez import models


def get_intensity_statistics(image: SimpleITK.Image, mask_image: SimpleITK.Image, model: models.Model, out_csv: str) -> None:
    """
    Get the intensity statistics of a NIFTI image file.

    :param image: The source image from which the intensity statistics are calculated.
    :type image: sitk.Image
    :param mask_image: The multilabel mask image.
    :type mask_image: sitk.Image
    :param model: The model.
    :type model: Model
    :param out_csv: The path to the output CSV file.
    :type out_csv: str
    :return: None
    """
    intensity_statistics = SimpleITK.LabelIntensityStatisticsImageFilter()
    min_intensity = SimpleITK.GetArrayViewFromImage(image).min()
    max_intensity = SimpleITK.GetArrayViewFromImage(image).max()
    bins = int(max_intensity - min_intensity)
    intensity_statistics.SetNumberOfBins(bins)
    intensity_statistics.Execute(mask_image, image)
    stats_list = [(intensity_statistics.GetMean(i), intensity_statistics.GetStandardDeviation(i),
                   intensity_statistics.GetMedian(i), intensity_statistics.GetMaximum(i),
                   intensity_statistics.GetMinimum(i)) for i in intensity_statistics.GetLabels()]
    columns = ['Mean', 'Standard-Deviation', 'Median', 'Maximum', 'Minimum']
    stats_df = pd.DataFrame(data=stats_list, index=intensity_statistics.GetLabels(), columns=columns)
    labels_present = stats_df.index.to_list()
    regions_present = []
    organ_indices_dict = model.organ_indices
    for label in labels_present:
        if label in organ_indices_dict:
            regions_present.append(organ_indices_dict[label])
        else:
            continue
    stats_df.insert(0, 'Regions-Present', np.array(regions_present))
    stats_df.to_csv(out_csv)


def get_shape_statistics(mask_image: SimpleITK.Image, model: models.Model, out_csv: str) -> None:
    """
    Get the shape statistics of a NIFTI image file.

    :param mask_image: The multilabel mask image.
    :type mask_image: sitk.Image
    :param model: The model.
    :type model: Model
    :param out_csv: The path to the output CSV file.
    :type out_csv: str
    :return: None
    """
    label_shape_filter = SimpleITK.LabelShapeStatisticsImageFilter()
    label_shape_filter.Execute(mask_image)

    stats_list = [(label_shape_filter.GetPhysicalSize(i),) for i in label_shape_filter.GetLabels() if
                  i != 0]  # exclude background label
    columns = ['Volume(mm3)']
    stats_df = pd.DataFrame(data=stats_list, index=[i for i in label_shape_filter.GetLabels() if i != 0],
                            columns=columns)

    labels_present = stats_df.index.to_list()
    regions_present = []
    organ_indices_dict = model.organ_indices
    for label in labels_present:
        if label in organ_indices_dict:
            regions_present.append(organ_indices_dict[label])
        else:
            continue
    stats_df.insert(0, 'Regions-Present', np.array(regions_present))
    stats_df.to_csv(out_csv)


def limit_fov(image_array: np.array, segmentation_array: np.array, fov_label: Union[List[int], int], largest_component_only: bool = False):

    if largest_component_only:
        segmentation_array = largest_connected_component(segmentation_array, fov_label)

    if isinstance(fov_label, list):
        z_indices = np.where((segmentation_array >= fov_label[0]) & (segmentation_array <= fov_label[1]))[0]
    else:
        z_indices = np.where(segmentation_array == fov_label)[0]
    z_min, z_max = np.min(z_indices), np.max(z_indices)

    # Crop the CT data along the z-axis
    limited_fov_array = image_array[z_min:z_max + 1, :, :]

    return limited_fov_array, {"z_min": z_min, "z_max": z_max, "original_shape": image_array.shape}


def expand_segmentation_fov(limited_fov_segmentation_array: np.ndarray, original_fov_info: Dict) -> np.ndarray:
    z_min = original_fov_info["z_min"]
    z_max = original_fov_info["z_max"]
    original_shape = original_fov_info["original_shape"]
    # Initialize an array of zeros with the shape of the original CT
    filled_segmentation_array = np.zeros(original_shape, np.uint8)
    # Place the cropped segmentation back into its original position
    filled_segmentation_array[z_min:z_max + 1, :, :] = limited_fov_segmentation_array

    return filled_segmentation_array


def largest_connected_component(segmentation_array, intensities):
    """
    Extracts the largest connected component for one or more specific intensities from a multilabel segmentation array
    and returns a new multilabel array where the largest components retain their original intensity.

    Parameters:
    - segmentation_array: 3D or 2D numpy array with multiple labels.
    - intensities: A single intensity or a list of intensities for which the largest component(s) should be extracted.

    Returns:
    - largest_components_multilabel: A multilabel array of the same shape as `segmentation_array`, where the largest
      connected component(s) of the specified intensity or intensities retain their original intensity, and all other
      areas are 0.
    """

    # Ensure intensities is a list (even if only one intensity is provided)
    if not isinstance(intensities, (list, tuple, np.ndarray)):
        intensities = [intensities]

    # Initialize an array to store the largest connected components
    largest_components_multilabel = np.zeros_like(segmentation_array, dtype=segmentation_array.dtype)

    # Loop over each intensity
    for intensity in intensities:
        # Create a binary mask for the current intensity
        binary_mask = segmentation_array == intensity

        # Label connected components in the binary mask
        labeled_array, num_features = ndimage.label(binary_mask)

        # Find the sizes of each connected component
        component_sizes = np.bincount(labeled_array.ravel())

        # Ignore the background (component 0)
        component_sizes[0] = 0

        # Find the largest connected component for this intensity
        largest_component_label = component_sizes.argmax()

        # Create a mask for the largest connected component of this intensity
        largest_component = labeled_array == largest_component_label

        # Assign the original intensity value to the largest connected component
        largest_components_multilabel[largest_component] = intensity

    return largest_components_multilabel


def extract_labels_and_write(multilabel_image: SimpleITK.Image, model: models.Model, output_directory: str):
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    label_stats = SimpleITK.LabelStatisticsImageFilter()
    label_stats.Execute(multilabel_image, multilabel_image)
    unique_labels = label_stats.GetLabels()

    for label in unique_labels:
        if label == 0:
            continue

        if label not in model.organ_indices:
            continue

        label_mask = SimpleITK.BinaryThreshold(multilabel_image, lowerThreshold=label, upperThreshold=label, insideValue=1, outsideValue=0)

        label_name = model.organ_indices[label]
        file_path = os.path.join(output_directory, f"{label_name}.nii.gz")
        SimpleITK.WriteImage(label_mask, file_path)


def image_get_orientation_code(image: SimpleITK.Image) -> str:
    image_orientation_code = SimpleITK.DICOMOrientImageFilter_GetOrientationFromDirectionCosines(image.GetDirection())
    return image_orientation_code


def image_reorient(image: SimpleITK.Image, new_orientation_code) -> SimpleITK.Image:
    image_orientation_code = image_get_orientation_code(image)
    if image_orientation_code != new_orientation_code:
        return SimpleITK.DICOMOrient(image, new_orientation_code)
    return image


def image_read(image_path: str) -> SimpleITK.Image:
    SimpleITK.ProcessObject_SetGlobalWarningDisplay(False)
    try:
        image = SimpleITK.ReadImage(image_path)
    except RuntimeError:
        with tempfile.TemporaryDirectory() as tmpdir:
            file_tmp_path = os.path.join(tmpdir, "non_orthonormal.nii.gz")
            image_nibabel = nibabel.load(image_path)
            nibabel.save(nibabel.Nifti1Image(image_nibabel.get_fdata(), image_nibabel.get_qform()), file_tmp_path)
            image = SimpleITK.ReadImage(file_tmp_path)
    SimpleITK.ProcessObject_SetGlobalWarningDisplay(True)
    return image


class ImageChunker:
    @staticmethod
    def __compute_interior_indices(axis_length: int, number_of_chunks: int) -> Tuple[List[int], List[int]]:
        chunk_base_size = axis_length // number_of_chunks
        remaining_slices = axis_length % number_of_chunks

        start_indices = []
        end_indices = []
        chunk_current_index = 0

        for i in range(number_of_chunks):
            chunk_size = chunk_base_size + (1 if i < remaining_slices else 0)
            start_indices.append(chunk_current_index)
            chunk_current_index += chunk_size
            end_indices.append(chunk_current_index)

        return start_indices, end_indices

    @staticmethod
    def __chunk_array_with_overlap(array_shape: Union[List[int], Tuple[int, ...]], splits_per_dimension: Union[List[int], Tuple[int, ...]], overlap_per_dimension: Union[List[int], Tuple[int, ...]]) -> List[Dict]:
        dims = array_shape
        num_dims = len(array_shape)
        starts_list = []
        ends_list = []

        for dimension_index in range(num_dims):
            axis_length = dims[dimension_index]
            number_of_chunks = splits_per_dimension[dimension_index]
            start_index, end_index = ImageChunker.__compute_interior_indices(axis_length, number_of_chunks)
            starts_list.append(start_index)
            ends_list.append(end_index)

        chunk_info = []
        for idx in itertools.product(*(range(len(s)) for s in starts_list)):
            chunk_slice = []
            interior_slice = []
            dest_slice = []
            for dimension_index, chunk_index in enumerate(idx):
                start_index = starts_list[dimension_index]
                end_index = ends_list[dimension_index]
                axis_length = dims[dimension_index]
                number_of_chunks = splits_per_dimension[dimension_index]
                overlap = overlap_per_dimension[dimension_index]

                start = max(0, start_index[chunk_index] - overlap if chunk_index > 0 else start_index[chunk_index])
                end = min(axis_length, end_index[chunk_index] + overlap if chunk_index < number_of_chunks - 1 else end_index[chunk_index])

                start_in_chunk = start_index[chunk_index] - start
                end_in_chunk = start_in_chunk + (end_index[chunk_index] - start_index[chunk_index])

                start_in_full = start_index[chunk_index]
                end_in_full = end_index[chunk_index]

                chunk_slice.append(slice(start, end))
                interior_slice.append(slice(start_in_chunk, end_in_chunk))
                dest_slice.append(slice(start_in_full, end_in_full))

            chunk_info.append({
                'chunk_slice': tuple(chunk_slice),
                'interior_slice': tuple(interior_slice),
                'dest_slice': tuple(dest_slice)
            })

        return chunk_info

    @staticmethod
    def array_to_chunks(image_array: np.ndarray, splits_per_dimension: Union[List[int], Tuple[int, ...]], overlap_per_dimension: Union[List[int], Tuple[int, ...]]) -> Tuple[List[np.ndarray], List[Dict]]:
        chunk_info = ImageChunker.__chunk_array_with_overlap(image_array.shape, splits_per_dimension, overlap_per_dimension)
        image_chunks = []
        positions = []

        for info in chunk_info:
            image_chunk = image_array[info['chunk_slice']]
            positions.append({
                'interior_slice': info['interior_slice'],
                'dest_slice': info['dest_slice']
            })
            image_chunks.append(image_chunk)

        return image_chunks, positions

    @staticmethod
    def chunks_to_array(image_chunks: List[np.ndarray], image_chunk_positions: List[Dict], final_shape: Union[List[int], Tuple[int, ...]]) -> np.ndarray:
        final_arr = np.empty(final_shape, dtype=image_chunks[0].dtype)
        for image_chunk, image_chunk_position in zip(image_chunks, image_chunk_positions):
            interior_region = image_chunk[image_chunk_position['interior_slice']]
            final_arr[image_chunk_position['dest_slice']] = interior_region

        return final_arr

    @staticmethod
    def determine_splits(image_array: np.ndarray) -> Tuple:
        image_shape = image_array.shape
        splits = []
        for axis in image_shape:
            if axis == 1:
                split = 1
            else:
                split = max(1, math.ceil(axis / CHUNK_THRESHOLD_INFERRING))
            splits.append(split)

        return tuple(splits)


class ImageResampler:
    @staticmethod
    def chunk_along_axis(axis: int) -> int:
        """
        Determines the maximum number of evenly-sized chunks that the axis can be split into.
        Each chunk is at least of size CHUNK_THRESHOLD.

        :param axis: Length of the axis.
        :type axis: int
        :return: The maximum number of evenly-sized chunks.
        :rtype: int
        :raises ValueError: If axis is negative or if CHUNK_THRESHOLD is less than or equal to 0.
        """
        # Check for negative input values
        if axis < 0:
            raise ValueError('Axis must be non-negative')

        if CHUNK_THRESHOLD_RESAMPLING <= 0:
            raise ValueError('CHUNK_THRESHOLD must be greater than 0')

        # If the axis is smaller than the threshold, it cannot be split into smaller chunks
        if axis < CHUNK_THRESHOLD_RESAMPLING:
            return 1

        # Determine the maximum number of chunks that the axis can be split into
        split = axis // CHUNK_THRESHOLD_RESAMPLING

        # Reduce the number of chunks until the axis is evenly divisible by split
        while axis % split != 0:
            split -= 1

        return split

    @staticmethod
    def resample_chunk_SimpleITK(image_chunk: da.array, input_spacing: Tuple, interpolation_method: int,
                                 output_spacing: Tuple, output_size: Tuple) -> da.array:
        """
        Resamples a dask array chunk.

        :param image_chunk: The chunk (part of an image) to be resampled.
        :type image_chunk: da.array
        :param input_spacing: The original spacing of the chunk (part of an image).
        :type input_spacing: tuple
        :param interpolation_method: SimpleITK interpolation type.
        :type interpolation_method: int
        :param output_spacing: Spacing of the newly resampled chunk.
        :type output_spacing: tuple
        :param output_size: Size of the newly resampled chunk.
        :type output_size: tuple
        :return: The resampled chunk (part of an image).
        :rtype: da.array
        """
        sitk_image_chunk = SimpleITK.GetImageFromArray(image_chunk)
        sitk_image_chunk.SetSpacing(input_spacing)

        resampled_sitk_image = SimpleITK.Resample(sitk_image_chunk, output_size, SimpleITK.Transform(),
                                                  interpolation_method, sitk_image_chunk.GetOrigin(), output_spacing,
                                                  sitk_image_chunk.GetDirection(), 0.0,
                                                  sitk_image_chunk.GetPixelIDValue())

        resampled_array = SimpleITK.GetArrayFromImage(resampled_sitk_image)
        return resampled_array

    @staticmethod
    def resample_image_SimpleITK_DASK_array(sitk_image: SimpleITK.Image, interpolation: str,
                                            output_spacing: Tuple[float, float, float] = (1.5, 1.5, 1.5),
                                            output_size: Union[Tuple[float, float, float], None] = None) -> np.array:
        if interpolation == 'nearest':
            interpolation_method = SimpleITK.sitkNearestNeighbor
        elif interpolation == 'linear':
            interpolation_method = SimpleITK.sitkLinear
        elif interpolation == 'bspline':
            interpolation_method = SimpleITK.sitkBSpline
        else:
            raise ValueError('The interpolation method is not supported.')

        input_spacing = sitk_image.GetSpacing()
        input_size = sitk_image.GetSize()
        input_chunks = [axis / ImageResampler.chunk_along_axis(axis) for axis in input_size]
        input_chunks_reversed = list(reversed(input_chunks))

        image_dask = da.from_array(SimpleITK.GetArrayViewFromImage(sitk_image), chunks=input_chunks_reversed)

        if output_size is not None:
            output_spacing = [input_spacing[i] * (input_size[i] / output_size[i]) for i in range(len(input_size))]

        output_chunks = [round(input_chunks[i] * (input_spacing[i] / output_spacing[i])) for i in
                         range(len(input_chunks))]
        output_chunks_reversed = list(reversed(output_chunks))

        result = da.map_blocks(ImageResampler.resample_chunk_SimpleITK, image_dask, input_spacing, interpolation_method,
                               output_spacing, output_chunks, chunks=output_chunks_reversed, meta=np.array(()),
                               dtype=np.float32)

        return result.compute()

    @staticmethod
    def reslice_identity(reference_image: SimpleITK.Image, moving_image: SimpleITK.Image,
                         output_image_path: Union[str, None] = None, is_label_image: bool = False) -> SimpleITK.Image:
        """
        Reslices an image to the same space as another image.

        :param reference_image: The reference image.
        :type reference_image: SimpleITK.Image
        :param moving_image: The image to reslice to the reference image.
        :type moving_image: SimpleITK.Image
        :param output_image_path: Path to the resliced image. Default is None.
        :type output_image_path: str
        :param is_label_image: Determines if the image is a label image. Default is False.
        :type is_label_image: bool
        :return: The resliced image as SimpleITK.Image.
        :rtype: SimpleITK.Image
        """
        resampler = SimpleITK.ResampleImageFilter()
        resampler.SetReferenceImage(reference_image)

        if is_label_image:
            resampler.SetInterpolator(SimpleITK.sitkNearestNeighbor)
        else:
            resampler.SetInterpolator(SimpleITK.sitkLinear)

        resampled_image = resampler.Execute(moving_image)
        resampled_image = SimpleITK.Cast(resampled_image, SimpleITK.sitkInt32)
        if output_image_path is not None:
            SimpleITK.WriteImage(resampled_image, output_image_path)
        return resampled_image

    @staticmethod
    def resample_segmentation(reference_image: SimpleITK.Image, segmentation_image: SimpleITK.Image):
        resampled_sitk_image = SimpleITK.Resample(segmentation_image, reference_image.GetSize(), SimpleITK.Transform(),
                                                  SimpleITK.sitkNearestNeighbor, reference_image.GetOrigin(),
                                                  reference_image.GetSpacing(), reference_image.GetDirection(), 0.0,
                                                  segmentation_image.GetPixelIDValue())
        return resampled_sitk_image



moose_to_snomed = {
    "heart_myocardium": {
        "ID": "74281007", # checked
        "name": "Myocardium structure" # checked
    },
    "heart_atrium_left": {
        "ID": "82471001", # checked
        "name": "Left atrial structure" # checked
    },
    "heart_atrium_right": {
        "ID": "73829009", # checked
        "name": "Right atrial structure" # checked
    },
    "heart_ventricle_left": {
        "ID": "87878005", # checked
        "name": "Left ventricular structure" # checked
    },
    "heart_ventricle_right": {
        "ID": "53085002", # checked
        "name": "Right ventricular structure" # checked
    },
    "aorta": {
        "ID": "15825003", # checked
        "name": "Aortic structure" # checked
    },
    "iliac_artery_left": {
        "ID": "721077009", # checked
        "name": "Structure of left iliac artery" # checked
    },
    "iliac_artery_right": {
        "ID": "721035009", # checked
        "name": "Structure of right iliac artery" # checked
    },
    "iliac_vena_left": {
        "ID": "244411005", # checked
        "name": "Iliac vein structure" # checked
    },
    "iliac_vena_right": {
        "ID": "244411005", # checked
        "name": "Iliac vein structure" # checked
    },
    "inferior_vena_cava": {
        "ID": "64131007", # checked
        "name": "Inferior vena cava structure" # checked
    },
    "portal_splenic_vein": {
        "ID": "35819009", # checked
        "name": "Structure of splenic vein" # checked
    },
    "pulmonary_artery": {
        "ID": "81040000", # checked
        "name": "Pulmonary artery structure" # checked
    },
    "colon": {
        "ID": "71854001", # checked
        "name": "Colon structure" # checked
    },
    "duodenum": {
        "ID": "38848004", # checked
        "name": "Duodenum structure" # checked
    },
    "esophagus": {
        "ID": "32849002", # checked
        "name": "Esophageal structure" # checked
    },
    "small_bowel": {
        "ID": "30315005", # checked
        "name": "Small intestinal structure" # checked
    },
    "autochthon_left": {
        "ID": "44947003", # checked, but might need reevaluation
        "name": "Structure of erector spinae muscle" # checked, but might need reevaluation
    },
    "autochthon_right": {
        "ID": "44947003", # checked, but might need reevaluation
        "name": "Structure of erector spinae muscle" # checked, but might need reevaluation
    },
    "gluteus_maximus_left": {
        "ID": "206007", # checked, but might need reevaluation
        "name": "Structure of gluteus maximus muscle" # checked, but might need reevaluation
    },
    "gluteus_maximus_right": {
        "ID": "206007", # checked, but might need reevaluation
        "name": "Structure of gluteus maximus muscle" # checked, but might need reevaluation
    },
    "gluteus_medius_left": {
        "ID": "78333006", # checked, but might need reevaluation
        "name": "Structure of gluteus medius muscle" # checked, but might need reevaluation
    },
    "gluteus_medius_right": {
        "ID": "78333006", # checked, but might need reevaluation
        "name": "Structure of gluteus medius muscle" # checked, but might need reevaluation
    },
    "gluteus_minimus_left": {
        "ID": "75297007", # checked, but might need reevaluation
        "name": "Structure of gluteus minimus muscle" # checked, but might need reevaluation
    },
    "gluteus_minimus_right": {
        "ID": "75297007", # checked, but might need reevaluation
        "name": "Structure of gluteus minimus muscle" # checked, but might need reevaluation
    },
    "iliopsoas_left": {
        "ID": "68455001", # checked, but might need reevaluation
        "name": "Structure of iliopsoas muscle" # checked, but might need reevaluation
    },
    "iliopsoas_right": {
        "ID": "68455001", # checked, but might need reevaluation
        "name": "Structure of iliopsoas muscle" # checked, but might need reevaluation
    },
    "adrenal_gland_left": {
        "ID": "12003004", # checked
        "name": "Left adrenal gland" # checked
    },
    "adrenal_gland_right": {
        "ID": "29392005", # checked
        "name": "Right adrenal gland" # checked
    },
    "bladder": {
        "ID": "89837001", # checked
        "name": "Urinary bladder structure" # checked
    },
    "brain": {
        "ID": "12738006", # checked
        "name": "Brain structure" # checked
    },
    "gallbladder": {
        "ID": "28231008", # checked
        "name": "Gallbladder structure" # checked
    },
    "kidney_left": {
        "ID": "18639004", # checked
        "name": "Left kidney structure" # checked
    },
    "kidney_right": {
        "ID": "9846003", # checked
        "name": "Right kidney structure" # checked
    },
    "liver": {
        "ID": "10200004", # checked
        "name": "Liver structure" # checked
    },
    "lung_lower_lobe_left": {
        "ID": "41224006", # checked
        "name": "Lower lobe of left lung" # checked
    },
    "lung_lower_lobe_right": {
        "ID": "266005", # checked
        "name": "Lower lobe of right lung" # checked
    },
    "lung_middle_lobe_right": {
        "ID": "72481006", # checked
        "name": "Middle lobe of right lung" # checked
    },
    "lung_upper_lobe_left": {
        "ID": "44714003", # checked
        "name": "Upper lobe of left lung" # checked
    },
    "lung_upper_lobe_right": {
        "ID": "42400003", # checked
        "name": "Upper lobe of right lung" # checked
    },
    "pancreas": {
        "ID": "15776009", # checked
        "name": "Pancreatic structure" # checked
    },
    "spleen": {
        "ID": "78961009", # checked
        "name": "Splenic structure" # checked
    },
    "stomach": {
        "ID": "69695003", # checked
        "name": "Stomach structure" # checked
    },
    "thyroid_left": {
        "ID": "79163004", # checked
        "name": "Left lobe of thyroid gland" # checked
    },
    "thyroid_right": {
        "ID": "29565003", # checked
        "name": "Right lobe of thyroid gland" # checked
    },
    "trachea": {
        "ID": "44567001", # checked
        "name": "Trachea structure" # checked
    },
    "carpal_left": {
        "ID": "719623009", # checked, but might need reevaluation
        "name": "Bone structure of left carpus" # checked, but might need reevaluation
    },
    "carpal_right": {
        "ID": "719624003", # checked, but might need reevaluation
        "name": "Bone structure of right carpus" # checked, but might need reevaluation
    },
    "clavicle_left": {
        "ID": "720617006", # checked
        "name": "Bone structure of left clavicle" # checked
    },
    "clavicle_right": {
        "ID": "720616002", # checked
        "name": "Bone structure of right clavicle" # checked
    },
    "femur_left": {
        "ID": "1356755002", # checked
        "name": "Entire bone of left femur" # checked
    },
    "femur_right": {
        "ID": "1356756001", # checked
        "name": "Entire bone of right femur" # checked
    },
    "fibula_left": {
        "ID": "1017225004", # checked
        "name": "Entire left fibula" # checked
    },
    "fibula_right": {
        "ID": "1017224000", # checked
        "name": "Entire right fibula" # checked
    },
    "fingers_left": {
        "ID": "762870007", # checked, but might need reevaluation
        "name": "Bone structure of phalanx of left hand" # checked, but might need reevaluation
    },
    "fingers_right": {
        "ID": "762871006", # checked, but might need reevaluation
        "name": "Bone structure of phalanx of right hand" # checked, but might need reevaluation
    },
    "humerus_left": {
        "ID": "719460003", # checked
        "name": "Bone structure of left humerus" # checked
    },
    "humerus_right": {
        "ID": "719461004", # checked
        "name": "Bone structure of right humerus" # checked
    },
    "metacarpal_left": {
        "ID": "762008002", # checked, but might need reevaluation
        "name": "Structure of metacarpal bone of left hand" # checked, but might need reevaluation
    },
    "metacarpal_right": {
        "ID": "762009005", # checked, but might need reevaluation
        "name": "Structure of metacarpal bone of right hand" # checked, but might need reevaluation
    },
    "metatarsal_left": {
        "ID": "726438004", # checked, but might need reevaluation
        "name": "Structure of metatarsal bone of left foot" # checked, but might need reevaluation
    },
    "metatarsal_right": {
        "ID": "726439007", # checked, but might need reevaluation
        "name": "Structure of metatarsal bone of right foot" # checked, but might need reevaluation
    },
    "patella_left": {
        "ID": "734208001", # checked
        "name": "Bone structure of left patella" # checked
    },
    "patella_right": {
        "ID": "734210004", # checked
        "name": "Bone structure of right patella" # checked
    },
    "radius_left": {
        "ID": "719464007", # checked
        "name": "Bone structure of left radius" # checked
    },
    "radius_right": {
        "ID": "719465008", # checked
        "name": "Bone structure of right radius" # checked
    },
    "scapula_left": {
        "ID": "719627005", # checked
        "name": "Bone structure of left scapula" # checked
    },
    "scapula_right": {
        "ID": "719628000", # checked
        "name": "Bone structure of right scapula" # checked
    },
    "skull": {
        "ID": "118646007", # checked
        "name": "Entire bone of head" # checked
    },
    "tarsal_left": {
        "ID": "108371006", # checked, but might need reevaluation
        "name": "Bone structure of tarsus" # checked, but might need reevaluation
    },
    "tarsal_right": {
        "ID": "108371006", # checked, but might need reevaluation
        "name": "Bone structure of tarsus" # checked, but might need reevaluation
    },
    "tibia_left": {
        "ID": "719492002", # checked
        "name": "Bone structure of left tibia" # checked
    },
    "tibia_right": {
        "ID": "719491009", # checked
        "name": "Bone structure of right tibia" # checked
    },
    "toes_left": {
        "ID": "720642002", # checked, but might need reevaluation
        "name": "Bone structure of phalanx of left foot" # checked, but might need reevaluation
    },
    "toes_right": {
        "ID": "720641009", # checked, but might need reevaluation
        "name": "Bone structure of phalanx of right foot" # checked, but might need reevaluation
    },
    "ulna_left": {
        "ID": "719462006", # checked
        "name": "Bone structure of left ulna" # checked
    },
    "ulna_right": {
        "ID": "719463001", # checked
        "name": "Bone structure of right ulna" # checked
    },
    "rib_left_1": {
        "ID": "1354709002", # checked
        "name": "Bone structure of left first rib" # checked
    },
    "rib_left_2": {
        "ID": "1354712004", # checked
        "name": "Bone structure of left second rib" # checked
    },
    "rib_left_3": {
        "ID": "1354715002", # checked
        "name": "Bone structure of left third rib" # checked
    },
    "rib_left_4": {
        "ID": "1354718000", # checked
        "name": "Bone structure of left fourth rib" # checked
    },
    "rib_left_5": {
        "ID": "1354724006", # checked
        "name": "Bone structure of left fifth rib" # checked
    },
    "rib_left_6": {
        "ID": "1354726008", # checked
        "name": "Bone structure of left sixth rib" # checked
    },
    "rib_left_7": {
        "ID": "1354728009", # checked
        "name": "Bone structure of left seventh rib" # checked
    },
    "rib_left_8": {
        "ID": "1354733008", # checked
        "name": "Bone structure of left eighth rib" # checked
    },
    "rib_left_9": {
        "ID": "1354735001", # checked
        "name": "Bone structure of left ninth rib" # checked
    },
    "rib_left_10": {
        "ID": "1354737009", # checked
        "name": "Bone structure of left tenth rib" # checked
    },
    "rib_left_11": {
        "ID": "1354742001", # checked
        "name": "Bone structure of left eleventh rib" # checked
    },
    "rib_left_12": {
        "ID": "1354744000", # checked
        "name": "Bone structure of left twelfth rib" # checked
    },
    "rib_left_13": {
        "ID": "205460009", # checked
        "name": "Accessory rib" # checked
    },
    "rib_right_1": {
        "ID": "1354710007", # checked
        "name": "Bone structure of right first rib" # checked
    },
    "rib_right_2": {
        "ID": "1354713009", # checked
        "name": "Bone structure of right second rib" # checked
    },
    "rib_right_3": {
        "ID": "1354716001", # checked
        "name": "Bone structure of right third rib" # checked
    },
    "rib_right_4": {
        "ID": "1354719008", # checked
        "name": "Bone structure of right fourth rib" # checked
    },
    "rib_right_5": {
        "ID": "1354725007", # checked
        "name": "Bone structure of right fifth rib" # checked
    },
    "rib_right_6": {
        "ID": "1354727004", # checked
        "name": "Bone structure of right sixth rib" # checked
    },
    "rib_right_7": {
        "ID": "1354729001", # checked
        "name": "Bone structure of right seventh rib" # checked
    },
    "rib_right_8": {
        "ID": "1354734002", # checked
        "name": "Bone structure of right eighth rib" # checked
    },
    "rib_right_9": {
        "ID": "1354736000", # checked
        "name": "Bone structure of right ninth rib" # checked
    },
    "rib_right_10": {
        "ID": "1354738004", # checked
        "name": "Bone structure of right tenth rib" # checked
    },
    "rib_right_11": {
        "ID": "1354743006", # checked
        "name": "Bone structure of right eleventh rib" # checked
    },
    "rib_right_12": {
        "ID": "1354745004", # checked
        "name": "Bone structure of right twelfth rib" # checked
    },
    "rib_right_13": {
        "ID": "205460009", # checked
        "name": "Accessory rib" # checked
    },
    "sternum": {
        "ID": "302522007", # checked
        "name": "Entire sternum" # checked
    },
    "vertebra_C1": {
        "ID": "731658004", # checked
        "name": "Entire bone of atlas" # checked
    },
    "vertebra_C2": {
        "ID": "731673003", # checked
        "name": "Entire bone of axis" # checked
    },
    "vertebra_C3": {
        "ID": "731721006", # checked
        "name": "Entire bone of C3" # checked
    },
    "vertebra_C4": {
        "ID": "731650006", # checked
        "name": "Entire bone of C4" # checked
    },
    "vertebra_C5": {
        "ID": "731670000", # checked
        "name": "Entire bone of C5" # checked
    },
    "vertebra_C6": {
        "ID": "731668009", # checked
        "name": "Entire bone of C6" # checked
    },
    "vertebra_C7": {
        "ID": "731715007", # checked
        "name": "Entire bone of C7" # checked
    },
    "vertebra_T1": {
        "ID": "731692001", # checked
        "name": "Entire bone of T1" # checked
    },
    "vertebra_T2": {
        "ID": "731683004", # checked
        "name": "Entire bone of T2" # checked
    },
    "vertebra_T3": {
        "ID": "731648003", # checked
        "name": "Entire bone of T3" # checked
    },
    "vertebra_T4": {
        "ID": "731700006", # checked
        "name": "Entire bone of T4" # checked
    },
    "vertebra_T5": {
        "ID": "731685006", # checked
        "name": "Entire bone of T5" # checked
    },
    "vertebra_T6": {
        "ID": "731676006", # checked
        "name": "Entire bone of T6" # checked
    },
    "vertebra_T7": {
        "ID": "731689000", # checked
        "name": "Entire bone of T7" # checked
    },
    "vertebra_T8": {
        "ID": "731654002", # checked
        "name": "Entire bone of T8" # checked
    },
    "vertebra_T9": {
        "ID": "731711003", # checked
        "name": "Entire bone of T9" # checked
    },
    "vertebra_T10": {
        "ID": "731652003", # checked
        "name": "Entire bone of T10" # checked
    },
    "vertebra_T11": {
        "ID": "731656000", # checked
        "name": "Entire bone of T11" # checked
    },
    "vertebra_T12": {
        "ID": "731662005", # checked
        "name": "Entire bone of T12" # checked
    },
    "vertebra_L1": {
        "ID": "731695004", # checked
        "name": "Entire bone of L1" # checked
    },
    "vertebra_L2": {
        "ID": "731657009", # checked
        "name": "Entire bone of L2" # checked
    },
    "vertebra_L3": {
        "ID": "731669001", # checked
        "name": "Entire bone of L3" # checked
    },
    "vertebra_L4": {
        "ID": "731655001", # checked
        "name": "Entire bone of L4" # checked
    },
    "vertebra_L5": {
        "ID": "731680001", # checked
        "name": "Entire bone of L5" # checked
    },
    "vertebra_L6": {
        "ID": "82137000", # checked
        "name": "Entire bone of L6" # checked
    },
    "hip_left": {
        "ID": "287679003", # checked
        "name": "Left hip region structure" # checked
    },
    "hip_right": {
        "ID": "287579007", # checked
        "name": "Right hip region structure" # checked
    },
    "sacrum": {
        "ID": "699698002", # checked
        "name": "Sacrum" # checked
    },
    "skeletal_muscle": {
        "ID": "127954009", # checked
        "name": "Skeletal muscle structure" # checked
    },
    "subcutaneous_fat": {
        "ID": "67769002", # checked
        "name": "Subcutaneous adipose tissue" # checked
    },
    "visceral_fat": {
        "ID": "725273006", # checked, but might need reevaluation
        "name": "Structure of adipose tissue of abdomen" # checked, but might need reevaluation
    }
}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This module contains the constants that are used in the moosez.

.. moduleauthor:: Lalith Kumar Shiyam Sundar <lalith.shiyamsundar@meduniwien.ac.at>
.. versionadded:: 3.0.0
"""

ALLOWED_MODALITIES = ['CT', 'PT', 'MR']

# COLOR CODES
ANSI_ORANGE = '\033[38;5;208m'
ANSI_GREEN = '\033[38;5;40m'
ANSI_VIOLET = '\033[38;5;141m'
ANSI_RED = '\033[38;5;196m'
ANSI_RESET = '\033[0m'

# FOLDER NAMES
SEGMENTATIONS_FOLDER = 'segmentations'
STATS_FOLDER = 'stats'

# PREPROCESSING PARAMETERS
INTERPOLATION = 'bspline'
CHUNK_THRESHOLD_RESAMPLING = 150
CHUNK_THRESHOLD_INFERRING = 350

# MODELS
KEY_FOLDER_NAME = "folder_name"
KEY_URL = "url"
KEY_LIMIT_FOV = "limit_fov"
KEY_DESCRIPTION = "description"
KEY_DESCRIPTION_TEXT = "Tissue of Interest"
KEY_DESCRIPTION_MODALITY = "Modality"
KEY_DESCRIPTION_IMAGING = "Imaging"
DEFAULT_SPACING = (1.5, 1.5, 1.5)
DEFAULT_TRANSPOSE_IDENTITY = [0, 1, 2]
OVERLAP_PER_AXIS = (0, 20, 20, 20)
FILE_NAME_DATASET_JSON = "dataset.json"
FILE_NAME_PLANS_JSON = "plans.json"


ENHANCE_URL = "https://enhance-pet-1-6k.s3.us-west-2.amazonaws.com/ENHANCE-PET-1_6k.zip"


USAGE_MESSAGE = """
Usage:
  moosez -d <MAIN_DIRECTORY> -m <MODEL_NAMES> -b
Example:  
  moosez -d /Documents/Data_to_moose/ -m clin_ct_organs

Description:
  MOOSE (Multi-organ objective segmentation) - A data-centric AI solution that
  generates multilabel organ segmentations for systemic TB whole-person research."""

"""

This module contains the constants that are used in the moosez.

Constants are values that are fixed and do not change during the execution of a program. 
They are used to store values that are used repeatedly throughout the program, such as 
file paths, folder names, and display parameters.

This module contains the following constants:

- `ALLOWED_MODALITIES`: A constant that stores a list of allowed modalities for the moosez algorithm.

- `ANSI_ORANGE`: A constant that stores the ANSI color code for orange.
- `ANSI_GREEN`: A constant that stores the ANSI color code for green.
- `ANSI_VIOLET`: A constant that stores the ANSI color code for violet.
- `ANSI_RESET`: A constant that stores the ANSI color code for resetting the color.

- `SEGMENTATIONS_FOLDER`: A constant that stores the name of the folder that contains the segmentations generated by 
                          the moosez algorithm.
- `STATS_FOLDER`: A constant that stores the name of the folder that contains the statistics generated by 
                  the moosez algorithm.

- `INTERPOLATION`: A constant that stores the interpolation method used by the moosez algorithm.
- `CHUNK_THRESHOLD`: A constant that stores the chunk threshold used by the moosez algorithm.

This module is imported by other modules in the moosez package and the constants are used throughout the package 
to provide fixed values that are used repeatedly.
"""
